/*=============================================================================
 * parser for CSP instances represented in XML format
 * 
 * Copyright (c) 2006 Olivier ROUSSEL (olivier.roussel <at> cril.univ-artois.fr)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *=============================================================================
 */
#ifndef _CSPVerifierCallback_h_
#define _CSPVerifierCallback_h_

#include <set>
#include <map>

#include "ExpressionParser.h"
#include "CSPParserCallback.h"

#define ShowNumericalIDS

namespace CSPXMLParser
{
  using namespace std;

  class CheckedConstraint: public VariableValuation
  {
  protected:
    // these vectors must have the same size
    vector<int> varValues;
    vector<bool> varAssigned; 

  public:
    void reserveFormalParameter(unsigned int pos)
    {
      if (pos>=varValues.size())
      {
	varValues.resize(pos+1);
	varAssigned.resize(pos+1);
      }
    }

    void clearVarValues()
    {
      for(unsigned int i=0;i<varAssigned.size();++i)
	varAssigned[i]=false;
    }

    void setVarValue(unsigned int pos, int value)
    {
      if (pos>=varValues.size())
      {
	cout << "ERROR: undefined parameter index (setVarValue) ("
	     << pos << ">=" << varValues.size() << ")" 
	     << endl;
	exit(1);
      }

      varValues[pos]=value;
      varAssigned[pos]=true;
    }

    virtual FunctionValue getVarValue(unsigned int id) const
    {
      if (id>=varValues.size())
      {
	cout << "ERROR: undefined parameter index (getVarValue)" << endl;
	exit(1);
      }

      if (!varAssigned[id])
      {
	cout << "ERROR: unassigned parameter" << endl;
	exit(1);
      }

      return varValues[id];
    }

    void dumpEffectiveParameters(ostream &s)
    {
      for(unsigned int i=0;i<varValues.size();++i)
      {
         if (varAssigned[i])
	   cout << "parm[" << i << "]=" << varValues[i] << endl;
	 else
	   cout << "parm[" << i << "]=<unassigned>" << endl;
      }
    }

    virtual void checkConstraint(const string &constraintName)=0;
  };


  class Domain
  {
  private:
    set<int> domain;
  public:
    void add(int val)
    {
      domain.insert(val);
    }

    void add(int first, int last)
    {
      for(int val=first;val<=last;++val)
	domain.insert(val);
    }

    bool belong(int val)
    {
      return domain.find(val)!=domain.end();
    }
  };

  class Predicate: public CheckedConstraint
  {
  private:
    AST *tree;

  public:
    Predicate()
    {
      tree=NULL;
    }

    void setAST(AST *tree)
    {
      this->tree=tree;
    }

    bool getValue()
    {
      return tree->value(*this);
    }

    void dump(ostream &s)
    {
      tree->expression(s);
    }

    virtual void checkConstraint(const string &constraintName)
    {
      if (!getValue())
      {
	cout << "ERROR: constraint " << constraintName 
	     << " is unsatisfied" << endl;

	cout << "Predicate is " << endl;
	dump(cout);
	cout << endl;

	dumpEffectiveParameters(cout);

	exit(1);
      }
    }
  };

  class Relation: public CheckedConstraint
  {
  private:
    unsigned int arity;
    unsigned int expectedNbTuples;
    bool isSupport;
    vector<vector<int> > tuples;

    bool tuplesAreSorted; // are tuples sorted ?

  public:
    void init(int arity, int nbTuples, bool isSupport) 
    {
      this->arity=arity;
      expectedNbTuples=nbTuples;
      this->isSupport=isSupport;
    }

    void addTuple(unsigned int arity, int tuple[]) 
    {
      if (arity!=this->arity)
      {
	cout << "ERROR: a tuple has an unexpected arity" << endl;
	exit(1);
      }

      vector<int> tmptuple;

      for(unsigned int i=0;i<arity;++i)
	tmptuple.push_back(tuple[i]);

      if (!tuples.empty())
	tuplesAreSorted=tuplesAreSorted && (tuples.back()<tmptuple);

      tuples.push_back(tmptuple);
    }

    void sortTuples()
    {
      sort(tuples.begin(),tuples.end());
    }

    void endTuples()
    {
      if (!tuplesAreSorted)
      {
	sortTuples();
	tuplesAreSorted=true;
      }
    }

    void dumpTuples(ostream &s)
    {
      s << "Relation contains ";
      for(unsigned int i=0;i<tuples.size();++i)
      {
	vector<int> &tuple=tuples[i];

	s << "(" << tuple[0];
	for(unsigned int j=1;j<tuple.size();++j)
	  s << "," << tuple[j];
	s << ")";
      }
      s << endl;
    }

    virtual void checkConstraint(const string &constraintName)
    {
      if (varAssigned.size()!=arity)
      {
	cout << "ERROR: number of assigned parameters differs from arity"
	     << " (relation check) in constraint"
	     << constraintName << endl;
	exit(1);
      }

      for(unsigned int i=0;i<arity;++i)
	if (!varAssigned[i])
	{
	  cout << "ERROR: unassigned parameter (relation check) in constraint "
	       << constraintName << endl;
	  exit(1);
	}

      bool found=binary_search(tuples.begin(),tuples.end(),varValues);

      if (found!=isSupport)
      {
	cout << "ERROR: relation in constraint "
	     << constraintName << " is not satisfied"
	     << endl;

	dumpEffectiveParameters(cout);

	if (isSupport)
	  cout << "Relation contains support tuples" << endl;
	else
	  cout << "Relation contains conflict tuples" << endl;

	dumpTuples(cout);

	exit(1);
      }
    }
  };

  class AllDifferent: public CheckedConstraint
  {
  public:
    virtual void checkConstraint(const string &constraintName)
    {
      set<int> S;

      if (varAssigned.size()==0)
      {
	cout << "ERROR: no variable in allDifferent constraint "
	     << constraintName << endl;
	exit(1);
      }


      for(unsigned int i=0;i<varAssigned.size();++i)
      {
	if (!varAssigned[i])
	{
	  cout << "ERROR: unassigned parameter in allDifferent constraint "
	       << constraintName << endl;
	  exit(1);
	}

	S.insert(varValues[i]);
      }

      if (S.size()!=varValues.size())
      {
 	cout << "ERROR: some variables are equal in allDifferent constraint "
	     << constraintName << endl;

	dumpEffectiveParameters(cout);
	exit(1);
      }
    }
  };

// definition of the functions which are called by the parser when it
// reads new information in the file. These functions are in charge of
// creating the data structure that the solver needs to do his job
  class CSPVerifierCallback : public CSPParserCallback
  {
  private:
    vector<int> varValues; // interpretation that must be checked

    unsigned int nbVar,scopePos;

    CheckedConstraint *currentConstraint;

    Domain *currentDomain;
    vector<Domain *> domainList;

    Relation *currentRelation;
    vector<Relation *> relationList;

    Predicate *currentPredicate;
    vector<Predicate *> predicateList;

    string currentConstraintName,currentConstraintReference;
    CSPDefinitionType currentConstraintType;

    map<string,CheckedConstraint *> globalConstraints;

  public:
    CSPVerifierCallback()
    {
      currentDomain=NULL;

      globalConstraints["global:allDifferent"]=new AllDifferent;
    }

    void clearVarValues()
    {
      varValues.clear();
    }

    void setVarValue(unsigned int i, int val)
    {
      if(i>=varValues.size())
	varValues.resize(i+1);

      varValues[i]=val;
    }

    virtual void beginInstance(const string & name)
    {
    }

    virtual void beginDomainsSection(int nbDomains) 
    {  
    }
  
    virtual void beginDomain(const string & name, int idDomain, int nbValue) 
    {
      currentDomain=new Domain;
      if (idDomain>=(int)domainList.size())
	domainList.resize(idDomain+1);

      domainList[idDomain]=currentDomain;
    }

    void addDomainValue(int v) 
    {
      currentDomain->add(v);
    }

    virtual void addDomainValue(int first,int last) 
    {
      currentDomain->add(first,last);      
    }

    virtual void endDomain() 
    {
    }

    /**
     * end the definition of all domains
     */
    virtual void endDomainsSection() 
    {
    }

    virtual void beginVariablesSection(int nbVariables) 
    {
      nbVar=0;
    }
  
    virtual void addVariable(const string & name, int idVar,
			     const string & domain, int idDomain) 
    {
      ++nbVar;

      if (idVar>=(int)varValues.size())
      {
	cout << "ERROR: variable " << name << " has no value!" << endl;
	exit(1);
      }

      if (!domainList[idDomain]->belong(varValues[idVar]))
      {
	cout << "ERROR: variable " << name << " has value " 
	     << varValues[idVar] 
	     << " which doesn't belong to its domain " << domain << endl;
	exit(1);
      }
    }

  virtual void endVariablesSection() 
  {
    if (nbVar!=varValues.size())
    {
      cout << "ERROR: the number of variables in the model is different from the number of variable in the instance!" << endl;
      exit(1);
    }
  }

  virtual void beginRelationsSection(int nbRelations) 
  {
  }

  
  virtual void beginRelation(const string & name, int idRel,
			     int arity, int nbTuples, bool isSupport) 
  {
    currentRelation=new Relation;
    if (idRel>=(int)relationList.size())
      relationList.resize(idRel+1);

    relationList[idRel]=currentRelation;

    currentRelation->init(arity,nbTuples,isSupport);
  }

  virtual void addRelationTuple(int arity, int tuple[]) 
  {
    currentRelation->addTuple(arity,tuple);
  }

  virtual void endRelation() 
  {
    currentRelation->endTuples();
    currentRelation=NULL;
  }

  virtual void endRelationsSection() 
  {
  }

  virtual void beginPredicatesSection(int nbPredicates) 
  {
  }
  
  virtual void beginPredicate(const string & name, int idPred) 
  {
    currentPredicate=new Predicate;
    if (idPred>=(int)predicateList.size())
      predicateList.resize(idPred+1);

    predicateList[idPred]=currentPredicate;
  }

  virtual void addFormalParameter(int pos, const string & name, 
				  const string & type) 
  {
    currentPredicate->reserveFormalParameter(pos);
  }

  virtual void predicateExpression(AST *tree) 
  {
    currentPredicate->setAST(tree);
  }

  virtual void predicateExpression(const string &expr) 
  {
    // we don't use this one
    throw runtime_error("unimplemented");
  }

  virtual void endPredicate() 
  {
    currentPredicate=NULL;
  }

  virtual void endPredicatesSection() 
  {
  }

  virtual void beginConstraintsSection(int nbConstraints) 
  {
  }
  
  virtual void beginConstraint(const string & name, int idConstr,
			       int arity, 
			       const string & reference, 
			       CSPDefinitionType type, int id)
  {
    currentConstraintName=name;
    currentConstraintReference=reference;
    currentConstraintType=type;

    switch(currentConstraintType)
    {
    case RelationType:
      currentConstraint=relationList[id];
      break;

    case PredicateType:
      currentConstraint=predicateList[id];
      break;

    case GlobalConstraintType:
      if (globalConstraints.find(reference)==
	  globalConstraints.end())
      {
	cout << "ERROR: unknown (or unimplemented) global constraint " 
	     << currentConstraintName << endl;
	exit(1);
      }

      currentConstraint=globalConstraints[reference];
      break;
    }

    currentConstraint->clearVarValues();
    scopePos=0;
  }

  virtual void addVariableToConstraint(const string & name, int idVar) 
  {
    switch(currentConstraintType)
    {
    case RelationType:
      currentConstraint->reserveFormalParameter(scopePos);
      currentConstraint->setVarValue(scopePos,varValues[idVar]);
      break;
    case GlobalConstraintType:
      currentConstraint->reserveFormalParameter(scopePos);
      if (currentConstraintReference=="global:allDifferent")
	currentConstraint->setVarValue(scopePos,varValues[idVar]);
      break;
    default:
      // don't care
      break;
    }
    
    scopePos++;
  }

  virtual void addEffectiveParameter(int pos, const string & expr) 
  {
    throw runtime_error("unimplemented");
  }

  virtual void addEffectiveParameter(int pos, AST *tree) 
  {
    throw runtime_error("unimplemented");
  }

  virtual void addEffectiveParameterVariable(int pos, 
					     const string & name, 
					     int idVar) 
  {
    if (currentConstraintType==PredicateType)
      currentConstraint->setVarValue(pos,varValues[idVar]);
  }

  virtual void addEffectiveParameterInteger(int pos, 
					    int value)
  {
    if (currentConstraintType==PredicateType)
      currentConstraint->setVarValue(pos,value);
  }

  virtual void addEffectiveParameterList(int pos) 
  {
    throw runtime_error("unimplemented");
  }

  /**
   * indicate the start of a list
   */
  virtual void beginList() 
  {
    throw runtime_error("unimplemented");
  }

  virtual void addListVariable(int pos, const string & name, int idVar) 
  {
    throw runtime_error("unimplemented");
  }

  virtual void addListInteger(int pos, int value) 
  {
    throw runtime_error("unimplemented");
  }

  virtual void endList() 
  {
    throw runtime_error("unimplemented");
  }

  virtual void endConstraint() 
  {
    currentConstraint->checkConstraint(currentConstraintName);
  }

  /**
   * end the definition of all constraints
   */
  virtual void endConstraintsSection() 
  {
  }

  /********************************************************************/


  /**
   * signal the end of parsing
   */
  virtual void endInstance() 
  {
  }

};

} // namespace

// Local Variables:
// mode: C++
// End:

#endif

