/*=============================================================================
 * parser for CSP instances represented in XML format
 * 
 * Copyright (c) 2006 Olivier ROUSSEL (olivier.roussel <at> cril.univ-artois.fr)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *=============================================================================
 */
#include <iostream>
#include <vector>
#include <getopt.h>

#include "CSPVerifierCallback.h"
#include "XMLParser.h"

using namespace CSPXMLParser;
using namespace std;

/**
 * read a line from the solver output (available from file f), skip
 * comments (lines starting with "c ") and store the first non comment
 * line in s
 */
void skipCommentsAndReadOneLine(istream &f, string &s)
{
  do
  {
    getline(f,s);
  }
  while(f.good() && s.substr(0,2)=="c ");
}


/**
 * read the interpretation
 */ 
void readInterpretation(istream &f, CSPVerifierCallback &cb)
{
  char c;
  int i,val;

  cb.clearVarValues();

  i=0;
  while(true)
  {
    c=f.peek();
    if (!f.good())
      break;

    if (c=='\r' || c=='\n')
      break;

    if (isspace(c))
    {
      f.get();
      continue;
    }

    f >> val;
    if (!f.good())
      break;

    cb.setVarValue(i,val);
    ++i;
  } 

  if (i==0)
  {
    cout << "ERROR: no interpretation found !" << endl;
    exit(7);
  }
}

/**
 * we expect a first line with SAT/UNSAT and then the certificate
 */
void readSolutionFileSyntax(CSPVerifierCallback &verifier, 
			    string expectedAnswer)
{
  string answer;
 
  skipCommentsAndReadOneLine(cin,answer);

  if (answer=="UNSAT")
  {
    if (expectedAnswer.length())
      if (expectedAnswer=="UNSAT")
      {
	cout << "OK" << endl;
	exit(0);
      }
      else
      {
	cout << "ERROR: got UNSAT answer instead of expected SAT answer" << endl;
	exit(1);
      }

    cout << "No possible verification on an UNSAT instance" << endl;
    exit(0);
  }
  else
    if (answer!="SAT")
    {
      cout << "ERROR: Unexpected answer ! (SAT/UNSAT expected)" << endl;
      cout << "Got answer: " << answer << endl;
      exit(1);
    }

    readInterpretation(cin,verifier);
}

/**
 * we expect a 's ' line and a 'v ' line
 */
void readStdoutSyntax(CSPVerifierCallback &verifier, 
		      string expectedAnswer)
{
  string answer,word;
 
  answer="<no 's ' line found>";

  while (cin.good())
  {
    cin >> word;
    if (!cin.good())
      break;

    if (word=="s")
    {
      cin >> answer;
      getline(cin,word);
    }
    else
      if (word=="v")
	readInterpretation(cin,verifier);
      else
	getline(cin,word);
  }

  if (answer=="UNSATISFIABLE")
  {
    if (expectedAnswer.length())
      if (expectedAnswer=="UNSAT")
      {
	cout << "OK" << endl;
	exit(0);
      }
      else
      {
	cout << "ERROR: got UNSATISFIABLE answer instead of expected SATISFIABLE answer" << endl;
	exit(1);
      }

    cout << "No possible verification on an UNSAT instance" << endl;
    exit(0);
  }
  else
    if (answer!="SATISFIABLE")
    {
      cout << "ERROR: Unexpected answer ! (SAT/UNSAT expected)" << endl;
      cout << "Got answer: " << answer << endl;
      exit(1);
    }
}

struct option longopts[] =
{
  {"solutionfile", no_argument, NULL, 'S'},
  {NULL, no_argument, NULL, 0}
};


int main(int argc, char **argv)
{
  CSPVerifierCallback verifier;
  bool solutionFileSyntax=false;

  int option;

  while ((option=getopt_long(argc,argv,"+",longopts,NULL))
	 !=EOF)
  {
    switch (option)
    {
    case 'S':
      solutionFileSyntax=true;
      break;
    }
  }

  if (argc!=2 && argc!=3)
  {
    cout << "usage: " << argv[0] << " [--solutionfile] <basename> [<expected answer>]" << endl
	 << "\tverifies the solution of a CSP instance" << endl
	 << "\tthe solution is read on the standard input" << endl
	 << "\tformula and solution must conform to the CSP'06 "
      "evaluation formats" << endl;
    cout << "\texpected answer can be SAT or UNSAT" << endl;
    
    exit(1);
  }

  string expectedAnswer;

  if (argc==3)
    expectedAnswer=argv[2];

  try
  {
    XMLParser<CSPVerifierCallback> parser(&verifier);

    parser.setPreferredExpressionRepresentation(AST::TREE);

    if (solutionFileSyntax)
      readSolutionFileSyntax(verifier,expectedAnswer);
    else
      readStdoutSyntax(verifier,expectedAnswer);

    int err=parser.parse(argv[1]); // parse the input file and verify

    // if the verifier callback didn't complain, then everything is OK
    if (err==0)
      cout << "OK" << endl;
    else
      cout << "ERROR: parser problems" << endl;
  }
  catch (exception &e)
  {
    cout.flush();
    cerr << "\n\tUnexpected exception :\n";
    cerr << "\t" << e.what() << endl;
    exit(1);
  }

  return 0;
}
