/*=============================================================================
 * parser for CSP instances represented in XML format
 * 
 * Copyright (c) 2006 Olivier ROUSSEL (olivier.roussel <at> cril.univ-artois.fr)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *=============================================================================
 */
#include "XMLParser.h"
#include "C_CSPParserCallback.h"

using namespace CSPXMLParser;

class C_Callback
{
private:
  void *solverData; // data passed by the solver to the callbacks
public:
  C_Callback(void *solverData)
  {
    this->solverData=solverData;
  }

  void beginInstance(const string & name) 
  {
    ::beginInstance(solverData,name.c_str());
  }

  void beginDomainsSection(int nbDomains)  
  {
    ::beginDomainsSection(solverData,nbDomains);
  }
  
  void beginDomain(const string & name, int idDomain, int nbValue)  
  {
    ::beginDomain(solverData,name.c_str(),idDomain,nbValue);
  }

  void addDomainValue(int v)  
  {
    ::addDomainValue(solverData,v);
  }

  void addDomainValue(int first,int last)  
  {
    ::addDomainValueInterval(solverData,first,last);
  }

  void endDomain()  
  {
    ::endDomain(solverData);
  }

  void endDomainsSection()  
  {
    ::endDomainsSection(solverData);
  }

  void beginVariablesSection(int nbVariables)  
  {
    ::beginVariablesSection(solverData,nbVariables);
  }
  
  void addVariable(const string & name, int idVar, 
			   const string & domain, int idDomain)  
  {
    ::addVariable(solverData,name.c_str(),idVar,domain.c_str(),idDomain);
  }

  void endVariablesSection()  
  {
    ::endVariablesSection(solverData);
  }


  void beginRelationsSection(int nbRelations)  
  {
    ::beginRelationsSection(solverData,nbRelations);
  }
  
  void beginRelation(const string & name, int idRel,
			     int arity, int nbTuples, bool isSupport)  
  {
    ::beginRelation(solverData,name.c_str(),idRel,arity,nbTuples,isSupport);
  }

  void addRelationTuple(int arity, int tuple[])  
  {
    ::addRelationTuple(solverData,arity,tuple);
  }

  void endRelation()  
  {
    ::endRelation(solverData);
  }

  void endRelationsSection()  
  {
    ::endRelationsSection(solverData);
  }

  void beginPredicatesSection(int nbPredicates)  
  {
    ::beginPredicatesSection(solverData,nbPredicates);
  }
  
  void beginPredicate(const string & name, int idPred)  
  {
    ::beginPredicate(solverData,name.c_str(),idPred);
  }

  void addFormalParameter(int pos, const string & name, const string & type)  
  {
    ::addFormalParameter(solverData,pos,name.c_str(),type.c_str());
  }

  void predicateExpression(AST *tree)  
  {
    ::predicateExpressionAST(solverData,NULL);
  }

  void predicateExpression(const string &expr)  
  {
    ::predicateExpression(solverData,expr.c_str()); 
  }

  void endPredicate()  
  {
    ::endPredicate(solverData);
  }

  void endPredicatesSection()  
  {
    ::endPredicatesSection(solverData);
  }

  void beginConstraintsSection(int nbConstraints)  
  {
    ::beginConstraintsSection(solverData,nbConstraints);
  }
  
  void beginConstraint(const string & name, int idConstr,
		       int arity, const string & reference, 
		       CSPDefinitionType type, int id)  
  {
    ::beginConstraint(solverData,name.c_str(),idConstr,arity,reference.c_str(),type,id);
  }

  void addVariableToConstraint(const string & name, int idVar)  
  {
    ::addVariableToConstraint(solverData,name.c_str(),idVar);
  }

  void addEffectiveParameter(int pos, const string & expr)  
  {
    ::addEffectiveParameter(solverData,pos,expr.c_str());
  }

  void addEffectiveParameter(int pos, AST *tree)  
  {
    ::addEffectiveParameterAST(solverData,pos,NULL);
  }

  void addEffectiveParameterVariable(int pos, const string & name, int idVar)  
  {
    ::addEffectiveParameterVariable(solverData,pos,name.c_str(),idVar);
  }

  void addEffectiveParameterInteger(int pos, int value)  
  {
    ::addEffectiveParameterInteger(solverData,pos,value);
  }

  void addEffectiveParameterList(int pos)  
  {
    ::addEffectiveParameterList(solverData,pos);
  }

  void beginList()  
  {
    ::beginList(solverData);
  }

  void addListVariable(int pos, const string & name, int idVar)  
  {
    ::addListVariable(solverData,pos,name.c_str(),idVar);
  }

  void addListInteger(int pos, int value)  
  {
    ::addListInteger(solverData,pos,value);
  }

  void endList()  
  {
    ::endList(solverData);
  }

  void endConstraint()  
  {
    ::endConstraint(solverData);
  }

  void endConstraintsSection()  
  {
    ::endConstraintsSection(solverData);
  }

  void endInstance()  
  {
    ::endInstance(solverData);
  }
};

void parseXML(char *filename, ExpressionSyntax preferredSyntax, 
	      void *solverData)
{
  C_Callback cb(solverData);

  try
  {
    XMLParser<C_Callback> parser(&cb);

    parser.setPreferredExpressionRepresentation((AST::Syntax)preferredSyntax);

    parser.parse(filename);
  }
  catch (exception &e)
  {
    cout.flush();
    cerr << "\n\tUnexpected exception :\n";
    cerr << "\t" << e.what() << endl;
    exit(1);
  }
}
